<?php
namespace Incoviba\Repository\Venta;

use PDO;
use PDOException;
use Incoviba\Common\Ideal;
use Incoviba\Common\Define;
use Incoviba\Common\Implement;
use Incoviba\Model;
use Incoviba\Service;

class Unidad extends Ideal\Repository
{
    public function __construct(Define\Connection $connection, protected Service\Proyecto\ProyectoTipoUnidad $proyectoTipoUnidadService)
    {
        parent::__construct($connection);
        $this->setTable('unidad');
    }

    public function create(?array $data = null): Model\Venta\Unidad
    {
        $map = (new Implement\Repository\MapperParser(['subtipo', 'piso', 'descripcion', 'orientacion', 'prorrateo']))
            ->register('pt', (new Implement\Repository\Mapper())
                ->setProperty('proyectoTipoUnidad')
                ->setFunction(function($data) {
                    return $this->proyectoTipoUnidadService->getById($data['pt']);
                }));
        return $this->parseData(new Model\Venta\Unidad(), $data, $map);
    }
    public function save(Define\Model $model): Model\Venta\Unidad
    {
        $model->id = $this->saveNew(
            ['subtipo', 'piso', 'descripcion', 'orientacion', 'pt'],
            [$model->subtipo, $model->piso, $model->descripcion, $model->orientacion, $model->proyectoTipoUnidad->id]
        );
        return $model;
    }
    public function edit(Define\Model $model, array $new_data): Model\Venta\Unidad
    {
        return $this->update($model, ['subtipo', 'piso', 'descripcion', 'orientacion', 'pt'], $new_data);
    }

    public function fetchById(int $id): Model\Venta\Unidad
    {
        $query = $this->connection->getQueryBuilder()
            ->select('a.*, up.prorrateo')
            ->from("{$this->getTable()} a")
            ->joined($this->joinProrrateo())
            ->where('a.id = ?');
        return $this->fetchOne($query, [$id]);
    }

    public function fetchByVenta(int $venta_id): array
    {
        $query = $this->connection->getQueryBuilder()
            ->select('a.*, up.prorrateo')
            ->from("{$this->getTable()} a")
            ->joined($this->joinProrrateo())
            ->joined('JOIN propiedad_unidad pu ON pu.unidad = a.id
    JOIN venta ON venta.propiedad = pu.propiedad')
            ->where('venta.id = ?');
        return $this->fetchMany($query, [$venta_id]);
    }
    public function fetchByPropiedad(int $propiedad_id): array
    {
        $query = $this->connection->getQueryBuilder()
            ->select('a.*, up.prorrateo')
            ->from("{$this->getTable()} a")
            ->joined($this->joinProrrateo())
            ->joined('JOIN `propiedad_unidad` pu ON pu.`unidad` = a.`id`')
            ->where('pu.propiedad = ?')
            ->group('a.id');
        return $this->fetchMany($query, [$propiedad_id]);
    }
    public function fetchByCierre(int $cierre_id): array
    {
        $query = $this->connection->getQueryBuilder()
            ->select('a.*, up.prorrateo')
            ->from("{$this->getTable()} a")
            ->joined($this->joinProrrateo())
            ->joined("JOIN `unidad_cierre` uc ON uc.`unidad` = a.`id`
    JOIN `proyecto_tipo_unidad` ptu ON ptu.`id` = a.`pt`
    JOIN `tipo_unidad` tu ON tu.`id` = ptu.`tipo`")
            ->where('uc.cierre = ?')
            ->group('a.id')
            ->order("tu.orden, LPAD(a.descripcion, 4, '0')");
        return $this->fetchMany($query, [$cierre_id]);
    }
    public function fetchByProyecto(int $proyecto_id): array
    {
        $query = $this->connection->getQueryBuilder()
            ->select('a.*, up.prorrateo')
            ->from("{$this->getTable()} a")
            ->joined($this->joinProrrateo())
            ->joined("INNER JOIN `proyecto_tipo_unidad` ptu ON ptu.`id` = a.`pt`")
            ->joined("INNER JOIN `tipo_unidad` tu ON tu.`id` = ptu.`tipo`")
            ->where('ptu.proyecto = ?')
            ->order('tu.orden');
        return $this->fetchMany($query, [$proyecto_id]);
    }
    public function fetchDisponiblesByProyecto(int $proyecto_id): array
    {
        $query = $this->connection->getQueryBuilder()
            ->select('DISTINCT a.*, up.prorrateo')
            ->from("{$this->getTable()} a")
            ->joined($this->joinProrrateo())
            ->joined('INNER JOIN `proyecto_tipo_unidad` ptu ON ptu.`id` = a.`pt`')
            ->joined('INNER JOIN `tipo_unidad` tu ON tu.`id` = ptu.`tipo`')
            ->joined('LEFT OUTER JOIN `propiedad_unidad` pu ON pu.`unidad` = a.`id`')
            ->joined('LEFT OUTER JOIN `venta` ON `venta`.`propiedad` = pu.`propiedad`')
            ->joined($this->joinEstadoVenta())
            ->joined('LEFT OUTER JOIN `tipo_estado_venta` tev ON tev.`id` = ev.`estado`')
            ->where("ptu.`proyecto` = ? AND (pu.`id` IS NULL OR `venta`.`id` IS NULL OR tev.`activa` = 0)")
            ->order('tu.orden');
        return $this->fetchMany($query, [$proyecto_id]);
    }
    public function fetchDisponiblesByDescripcionAndTipo(string $descripcion, string $tipo): array
    {
        $query = $this->connection->getQueryBuilder()
            ->select('DISTINCT a.*, up.prorrateo')
            ->from("{$this->getTable()} a")
            ->joined($this->joinProrrateo())
            ->joined('INNER JOIN proyecto_tipo_unidad ptu ON ptu.id = a.pt')
            ->joined('INNER tipo_unidad tu ON tu.id = ptu.tipo')
            ->joined('LEFT OUTER JOIN propiedad_unidad pu ON pu.unidad = a.id')
            ->joined('LEFT OUTER JOIN venta ON venta.propiedad = pu.propiedad')
            ->joined($this->joinEstadoVenta())
            ->joined('LEFT OUTER JOIN tipo_estado_venta tev ON tev.id = ev.estado')
            ->where("a.`descripcion` LIKE ? AND tu.`descripcion` = ? AND (pu.`id` IS NULL OR `venta`.`id` IS NULL OR tev.`activa` = 0)");
        return $this->fetchMany($query, [$descripcion, $tipo]);
    }

    /**
     * @param string $descripcion
     * @param string $tipo
     * @return array
     * @throws Implement\Exception\EmptyResult
     */
    public function fetchDisponiblesIdsByDescripcionAndTipo(string $descripcion, string $tipo): array
    {
        $query = $this->connection->getQueryBuilder()
            ->select('a.id')
            ->from("{$this->getTable()} a")
            ->joined('INNER JOIN `proyecto_tipo_unidad` ptu ON ptu.`id` = a.`pt`')
            ->joined('INNER JOIN `tipo_unidad` tu ON tu.`id` = ptu.`tipo`')
            ->joined('LEFT OUTER JOIN `propiedad_unidad` pu ON pu.`unidad` = a.`id`')
            ->joined('LEFT OUTER JOIN `venta` ON `venta`.`propiedad` = pu.`propiedad`')
            ->joined($this->joinEstadoVenta())
            ->joined('LEFT OUTER JOIN `tipo_estado_venta` tev ON tev.`id` = ev.`estado`')
            ->where("a.`descripcion` LIKE ? AND tu.`descripcion` = ? AND (pu.`id` IS NULL OR `venta`.`id` IS NULL OR tev.`activa` = 0)");
        try {
            return $this->connection->execute($query, [$descripcion, $tipo])->fetchAll(PDO::FETCH_ASSOC);
        } catch (PDOException $exception) {
            throw new Implement\Exception\EmptyResult($query, $exception);
        }
    }
    public function fetchByIdForSearch(int $unidad_id): array
    {
        $subSubQuery = $this->connection->getQueryBuilder()
            ->select('MAX(id) AS id, precio')
            ->from('estado_precio')
            ->group('precio');
        $subQuery = $this->connection->getQueryBuilder()
            ->select('ep1.*')
            ->from('estado_precio ep1')
            ->joined("INNER JOIN ($subSubQuery) ep0 ON ep0.id = ep1.id");
        $query = $this->connection->getQueryBuilder()
            ->select('unidad.id AS id, unidad.descripcion AS descripcion')
            ->columns('proyecto.id AS proyecto_id, proyecto.descripcion AS proyecto_descripcion')
            ->columns('tu.descripcion AS tipo_unidad_descripcion')
            ->columns('ptu.m2 + ptu.logia + ptu.terraza AS superficie')
            ->columns('precio.valor AS precio')
            ->from($this->getTable())
            ->joined('INNER JOIN proyecto_tipo_unidad ptu ON ptu.id = unidad.pt')
            ->joined('INNER JOIN proyecto ON proyecto.id = ptu.proyecto')
            ->joined('INNER JOIN tipo_unidad tu ON tu.id = ptu.tipo')
            ->joined('LEFT OUTER JOIN precio ON precio.unidad = unidad.id')
            ->joined("LEFT OUTER JOIN ({$subQuery}) ep ON ep.precio = precio.id")
            ->where('unidad.id = ?')
            ->group('unidad.id');
        return $this->connection->execute($query, [$unidad_id])->fetch(PDO::FETCH_ASSOC);
    }

    /**
     * @param int $unidad_id
     * @return bool
     * @throws Implement\Exception\EmptyResult
     */
    public function fetchSoldByUnidad(int $unidad_id): bool
    {
        $query = $this->connection->getQueryBuilder()
            ->select('a.id')
            ->from("{$this->getTable()} a")
            ->joined('INNER JOIN `propiedad_unidad` pu ON pu.`unidad` = a.`id`')
            ->joined('INNER JOIN `venta` ON `venta`.`propiedad` = `pu`.`propiedad`')
            ->joined($this->joinEstadoVenta())
            ->joined('LEFT OUTER JOIN `tipo_estado_venta` tev ON tev.`id` = ev.`estado`')
            ->where('a.id = :unidad_id AND tev.activa = 1');
        try {
            $result = $this->connection->execute($query, ['unidad_id' => $unidad_id])->fetch(PDO::FETCH_ASSOC);
            return $result !== false;
        } catch (PDOException $exception) {
            throw new Implement\Exception\EmptyResult($query, $exception);
        }
    }

    /**
     * @param array $unidad_ids
     * @return array
     * @throws Implement\Exception\EmptyResult
     */
    public function fetchSoldByUnidades(array $unidad_ids): array
    {
        $interrogations = implode(',', array_fill(0, count($unidad_ids), '?'));
        $query = $this->connection->getQueryBuilder()
            ->select('a.id')
            ->from("{$this->getTable()} a")
            ->joined('INNER JOIN `propiedad_unidad` pu ON pu.`unidad` = a.`id`')
            ->joined('INNER JOIN `venta` ON `venta`.`propiedad` = `pu`.`propiedad`')
            ->joined($this->joinEstadoVenta(true))
            ->joined('INNER JOIN `tipo_estado_venta` tev ON tev.`id` = ev.`estado`')
            ->where("a.id IN ({$interrogations}) AND tev.activa = 1");
        try {
            $results = $this->connection->execute($query, $unidad_ids)->fetchAll(PDO::FETCH_ASSOC);
            if (empty($results)) {
                throw new Implement\Exception\EmptyResult($query);
            }
            $missing = array_diff(array_map(fn($id) => (int) $id, $unidad_ids), array_column($results, 'id'));
            return [...array_map(function($result) {
                return [
                    'id' => $result['id'],
                    'sold' => true
                ];
            }, $results), ...array_map(function($id) {
                return [
                    'id' => $id,
                    'sold' => false
                ];
            }, $missing)];
        } catch (PDOException $exception) {
            throw new Implement\Exception\EmptyResult($query, $exception);
        }
    }

    /**
     * @param int $promotion_id
     * @return array
     * @throws Implement\Exception\EmptyResult
     */
    public function fetchByPromotion(int $promotion_id): array
    {
        $query = $this->connection->getQueryBuilder()
            ->select('a.*')
            ->from("{$this->getTable()} a")
            ->joined('INNER JOIN `promotion_units` pu ON pu.`unit_id` = a.`id`')
            ->where('pu.`promotion_id` = :promotion_id');
        return $this->fetchMany($query, ['promotion_id' => $promotion_id]);
    }

    /**
     * @param int $proyecto_id
     * @param int $tipo_id
     * @return array
     * @throws Implement\Exception\EmptyResult
     */
    public function fetchByProyectoAndTipo(int $proyecto_id, int $tipo_id): array
    {
        $query = $this->connection->getQueryBuilder()
            ->select('a.*')
            ->from("{$this->getTable()} a")
            ->joined('INNER JOIN proyecto_tipo_unidad ptu ON a.pt = ptu.id')
            ->where('ptu.`proyecto` = :proyecto_id AND ptu.`tipo` = :tipo_id');
        return $this->fetchMany($query, ['proyecto_id' => $proyecto_id, 'tipo_id' => $tipo_id]);
    }

    protected function joinProrrateo(): string
    {
        return "LEFT OUTER JOIN unidad_prorrateo up ON up.unidad_id = a.id";
    }
    protected function joinEstadoVenta(bool $inner = false): string
    {
        $subSubQuery = $this->connection->getQueryBuilder()
            ->select('MAX(id) AS id, venta')
            ->from('estado_venta')
            ->group('venta');
        $join = $inner ? 'INNER' : 'LEFT OUTER';
        $subQuery = $this->connection->getQueryBuilder()
            ->select('ev1.*')
            ->from('estado_venta ev1')
            ->joined("{$join} JOIN ($subSubQuery) ev0 ON ev0.id = ev1.id");
        return "{$join} JOIN ({$subQuery}) ev ON ev.`venta` = `venta`.`id`";
    }
}
